#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Unified Chip Design Platform - AMBA - AHB Drivers.
"""

from collections.abc import Iterable
from enum import IntEnum
from logging import getLogger

from cocotb.handle import SimHandle
from cocotb.triggers import RisingEdge


class BurstType(IntEnum):
    """HBURST encoding as per AHB spec."""

    SINGLE = 0  # Single transfer
    INCR = 1  # Unknown length burst
    WRAP4 = 2  # Four-beat wrap burst
    INCR4 = 3  # Four-beat incrementing burst
    WRAP8 = 4  # Eight-beat wrap burst
    INCR8 = 5  # Eight-beat incrementing burst
    WRAP16 = 6  # Sixteen-beat wrap burst
    INCR16 = 7  # Sixteen-beat incrementing burst


class TransType(IntEnum):
    """HTRANS encoding as per AHB spec."""

    IDLE = 0  # No transfer
    BUSY = 1  # Busy transfer
    NONSEQ = 2  # Non-sequential transfer
    SEQ = 3  # Sequential transfer


class SizeType(IntEnum):
    """HSIZE encoding as per AHB spec."""

    BYTE = 0  # 8-bit
    HALFWORD = 1  # 16-bit
    WORD = 2  # 32-bit
    DOUBLEWORD = 3  # 64-bit
    WORD4 = 4  # 128-bit
    WORD8 = 5  # 256-bit
    WORD16 = 6  # 512-bit
    WORD32 = 7  # 1024-bit


def _prep_addr_iter(addr: int, burst_length: int, size: SizeType, burst_type=BurstType) -> tuple[int, int, int, int]:
    """Prepare Address Iterations."""
    match burst_type:
        case BurstType.INCR16 | BurstType.WRAP16:
            mask = (1 << (size + 4)) - 1
            len = 16
        case BurstType.INCR8 | BurstType.WRAP8:
            mask = (1 << (size + 3)) - 1
            len = 8
        case BurstType.INCR4 | BurstType.WRAP4:
            mask = (1 << (size + 2)) - 1
            len = 4
        case BurstType.INCR:
            mask = -1
            len = burst_length
        case BurstType.SINGLE:
            mask = -1
            len = 1
    base = addr & ~mask
    offs = addr & mask
    return (base, offs, mask, len)


def _check_bus_acc(data_width: int, addr: int, offs: int, size: SizeType, burst_type=BurstType) -> None:
    """Check AHB Bus Access."""
    assert data_width >= (8 << size), f"Size argument {size!r} -> {8<<size} too big for data width of {data_width}!"
    assert (addr & ((1 << size) - 1)) == 0, f"Address {addr:x} is not aligned to size argument {size!r}!"
    assert (burst_type not in (BurstType.INCR16, BurstType.INCR8, BurstType.INCR4)) or (
        offs == 0
    ), f"Address {addr:x} is not aligned to BurstType {burst_type!r} at size {size!r}!"


class SlaveFsmState(IntEnum):
    """Internal AHB Slave State."""

    IDLE = 0
    ACTIVE = 1


class AHBMasterDriver:
    """AHB Master bus driver."""

    def __init__(
        self,
        name: str,
        clk: SimHandle,
        rst_an: SimHandle,
        haddr: SimHandle,
        hwrite: SimHandle,
        hwdata: SimHandle,
        htrans: SimHandle,
        hburst: SimHandle,
        hsize: SimHandle,
        hprot: SimHandle,
        hrdata: SimHandle,
        hready: SimHandle,
        hresp: SimHandle,
        hsel: SimHandle = None,
        log_level: int | None = None,
    ):
        self.name = name
        self.clk = clk
        self.rst_an = rst_an
        self.haddr = haddr
        self.hwrite = hwrite
        self.hwdata = hwdata
        self.htrans = htrans
        self.hburst = hburst
        self.hsize = hsize
        self.hprot = hprot
        self.hrdata = hrdata
        self.hready = hready
        self.hresp = hresp
        self.hsel = hsel  # allowed to be None as it might not be present (e.g. Multilayer input)
        self.data_width = len(hwdata)

        self.logger = getLogger(name)
        if log_level is not None:  # important explicit check for None as 0 would be a valid value
            self.logger.setLevel(log_level)

    async def write(
        self,
        addr: int,
        data: int | Iterable,
        size: SizeType = SizeType.WORD,
        burst_length: int = 1,
        burst_type: BurstType = BurstType.SINGLE,
    ) -> None:
        """AHB Write (Burst)."""
        base, offs, mask, burst_length = _prep_addr_iter(
            addr=addr, burst_length=burst_length, size=size, burst_type=burst_type
        )
        _check_bus_acc(data_width=self.data_width, addr=addr, offs=offs, size=size, burst_type=burst_type)

        if isinstance(data, int):
            log_data = (data,)
            data = iter((data,))
        else:
            log_data = tuple(data)
            data = iter(data)
        shmsk = self.data_width - 1
        self.haddr.value = base + offs
        self.hwdata.value = 0xDEADDEAD
        self.hwrite.value = 1
        if self.hsel:
            self.hsel.value = 1
        self.htrans.value = TransType.NONSEQ
        self.hburst.value = burst_type
        self.hsize.value = size
        await RisingEdge(self.clk)
        for _ in range(burst_length - 1):
            self.htrans.value = TransType.SEQ
            self.hwdata.value = next(data) << ((offs << 3) & shmsk)
            offs = (offs + (1 << size)) & mask
            self.haddr.value = base + offs
            await RisingEdge(self.clk)
            while self.hready == 0:
                await RisingEdge(self.clk)
        if self.hsel:
            self.hsel.value = 0
        self.haddr.value = 0
        self.hwdata.value = next(data) << ((offs << 3) & shmsk)
        self.hwrite.value = 0
        self.htrans.value = TransType.IDLE
        self.hburst.value = BurstType.SINGLE
        self.hsize.value = SizeType.BYTE
        await RisingEdge(self.clk)
        while self.hready == 0:
            await RisingEdge(self.clk)
        self.hwdata.value = 0xDEADDEAD
        hexwidth = (2**size) * 2
        self.logger.info(
            f"=MST WRITE= data: [{','.join(f'0x{x:0{hexwidth}X}' for x in log_data)}] "
            f"address: {hex(addr)} burst: {burst_type.name} burst length: {burst_length}  size: {size.name}"
        )

    async def read(
        self, addr: int, burst_length: int = 1, size: SizeType = SizeType.WORD, burst_type: BurstType = BurstType.SINGLE
    ) -> tuple[int]:
        """AHB Read (Burst)."""
        base, offs, mask, burst_length = _prep_addr_iter(
            addr=addr, burst_length=burst_length, size=size, burst_type=burst_type
        )
        _check_bus_acc(data_width=self.data_width, addr=addr, offs=offs, size=size, burst_type=burst_type)

        rdata = []
        poffs = offs
        shmsk = self.data_width - 1
        szmsk = (1 << (8 << size)) - 1
        self.haddr.value = base + offs
        if self.hsel:
            self.hsel.value = 1
        self.hwrite.value = 0
        self.htrans.value = TransType.NONSEQ
        self.hburst.value = burst_type
        self.hsize.value = size
        await RisingEdge(self.clk)

        for _ in range(burst_length - 1):
            self.htrans.value = TransType.SEQ
            offs = (offs + (1 << size)) & mask
            self.haddr.value = base + offs
            await RisingEdge(self.clk)
            while self.hready == 0:
                await RisingEdge(self.clk)
            rdata.append((self.hrdata.value.integer >> ((poffs << 3) & shmsk)) & szmsk)
            poffs = offs
        if self.hsel:
            self.hsel.value = 0
        self.haddr.value = 0
        self.htrans.value = TransType.IDLE
        self.hsize.value = SizeType.BYTE
        await RisingEdge(self.clk)
        while self.hready == 0:
            await RisingEdge(self.clk)
        rdata.append((self.hrdata.value.integer >> ((poffs << 3) & shmsk)) & szmsk)
        self.logger.info(
            f"=MST READ= data: [{','.join(f'0x{x:0{self.data_width // 4}X}' for x in rdata)}] "
            f"address: {hex(addr)} burst: {burst_type.name} burst length: {burst_length}  size: {size.name}"
        )
        return tuple(rdata)

    async def reset(self):
        """Reset AHB Master."""
        if self.hsel:
            self.hsel.value = 0
        self.hwrite.value = 0
        self.hwdata.value = 0
        self.htrans.value = 0  # IDLE
        self.hburst.value = 0
        self.hprot.value = 0


class AHBSlaveDriver:
    """Active AHB Slave that can respond to Master requests."""

    def __init__(
        self,
        name: str,
        clk: SimHandle,
        rst_an: SimHandle,
        haddr: SimHandle,
        hwrite: SimHandle,
        hwdata: SimHandle,
        htrans: SimHandle,
        hburst: SimHandle,
        hsize: SimHandle,
        hrdata: SimHandle,
        hready: SimHandle,
        hreadyout: SimHandle,
        hresp: SimHandle,
        hsel: SimHandle,
        hprot: SimHandle = None,
        hreadyout_delay: int = 0,
        size_bytes: int = 1024,
        log_level: int | None = None,
    ):
        """AHB Slave Init."""
        self.name = name
        self.clk = clk
        self.rst_an = rst_an
        self.haddr = haddr
        self.hwrite = hwrite
        self.hwdata = hwdata
        self.htrans = htrans
        self.hburst = hburst
        self.hsize = hsize
        self.hrdata = hrdata
        self.hready = hready
        self.hreadyout = hreadyout
        self.hresp = hresp
        self.hsel = hsel
        self.hprot = hprot  # allowed to be None as it might not be present/unsupported by a slave
        self.data_width = len(hwdata)

        self.logger = getLogger(name)
        if log_level is not None:  # important explicit check for None as 0 would be a valid value
            self.logger.setLevel(log_level)

        self.mem = bytearray(size_bytes)  # Initialize a 1KB memory
        self.hreadyout_delay = hreadyout_delay  # Delay for HREADYOUT signal to simulate longer access times
        self.addrmask = size_bytes - 1

        # state variables
        self.state = 0
        self.burst_count = 0  # Burst count for burst transactions
        self.curr_addr = None
        self.curr_wdata = None
        self.curr_write = None
        self.curr_size = None

    def read(self, addr, size):
        """AHB Read."""
        # number of bytes in this transfer according to transfer size
        byte_cnt = 2**size
        # extract the data from the bus
        alignmask = byte_cnt - 1
        shmsk = self.data_width - 1
        datashift_bit = (addr.integer << 3) & shmsk

        self.logger.debug(
            f"=SLV READ= alignment mask: {hex(alignmask)} "
            f"shift mask: {hex(shmsk)} datashift in bits: {hex(datashift_bit)}"
        )

        unaligned = alignmask & addr.integer
        assert not unaligned, f"Address is unaligned for write with HSIZE of {size} at HADDR {addr}."

        masked_addr = self.addrmask & addr.integer
        rdata = int.from_bytes(self.mem[masked_addr : masked_addr + byte_cnt], "little") << datashift_bit

        self.logger.info(
            f"=SLV READ= data: {hex(rdata)} address: {hex(addr.integer)} "
            f"address (masked): {hex(masked_addr)} byte count: {byte_cnt}"
        )

        return rdata

    def write(self, addr, size, data):
        """AHB Write."""
        # number of bytes in this transfer according to transfer size
        byte_cnt = 2**size
        shmsk = self.data_width - 1
        # extract the data from the bus
        alignmask = byte_cnt - 1
        # lower_addrmask = (byte_cnt * 2) - 1
        lower_datamask = (2 ** (byte_cnt * 8)) - 1
        # datashift_byte = addr.integer & lower_addrmask
        # datashift_bit = datashift_byte * 8
        datashift_bit = (addr.integer << 3) & shmsk
        unaligned = alignmask & addr.integer
        assert not unaligned, f"Address is unaligned for write with HSIZE of {size} at HADDR {addr}."

        self.logger.debug(
            f"=SLV WRITE= alignment mask: {hex(alignmask)} shift mask: {hex(shmsk)} "
            f"lower data mask {hex(lower_datamask)} datashift in bits: {hex(datashift_bit)}"
        )

        wdata = (data.integer >> datashift_bit) & lower_datamask
        masked_addr = self.addrmask & addr.integer
        bytes = int.to_bytes(wdata, byte_cnt, "little")

        self.logger.info(
            f"=SLV WRITE= data: {hex(wdata)} data (bytes): {','.join([hex(x) for x in bytes])} "
            f"address: {hex(addr.integer)} address (masked): {hex(masked_addr)} byte count: {byte_cnt}"
        )

        self.mem[masked_addr : masked_addr + byte_cnt] = bytes

    async def run(self):
        """Slave Main Loop."""
        self.hreadyout.value = 1
        self.hrdata.value = 0xDEADDEAD
        while True:
            await RisingEdge(self.clk)
            # print("BOZO", self.state, self.htrans.value, self.haddr.value)
            if self.state:
                for _ in range(self.hreadyout_delay):  # delay the answer if configured
                    await RisingEdge(self.clk)
                    self.hreadyout.value = 0
                self.hreadyout.value = 1
                self.curr_wdata = self.hwdata.value if self.curr_write else 0
                if self.curr_write:
                    # Handle write request
                    self.write(self.curr_addr, self.curr_size, self.curr_wdata)
            # Check if there's an AHB request
            if self.hsel.value and self.htrans.value in (TransType.SEQ, TransType.NONSEQ):
                self.curr_addr = self.haddr.value
                self.curr_write = self.hwrite.value
                self.curr_size = self.hsize.value
                self.state = 1
            else:
                self.state = 0
            if self.state and not self.curr_write:
                # Handle read request (need to apply read value for next cycle)
                rdata = self.read(self.curr_addr, self.curr_size)
                self.hrdata.value = rdata
            else:
                self.hrdata.value = 0xDEADBEEF  # BOZOjust temp for debugging
            # print("BOZO2", self.state)

    def set_hreadyout_delay(self, delay):
        """Set hreadyout Delay."""
        self.hreadyout_delay = delay

    def set_data(self, data):
        """Preload Slave Memory."""
        self.mem = data

    def log_data(self, start_addr=0, end_addr=None, chunk_size=16):
        """Request logging of Slave Memory content."""
        if end_addr is None:
            end_addr = len(self.mem) - 1
        elif end_addr > len(self.mem) - 1:
            self.logger.error(
                f"=MEMORY CONTENTS= Provided end_addr {hex(end_addr)} to log_data is beyond size of slave memory."
            )
            return
        for i in range(start_addr, end_addr + 1, chunk_size):
            numlen = len(f"{end_addr:X}")
            self.logger.info(
                f"=MEMORY CONTENTS= 0x{i:0{numlen}X}-0x{i+chunk_size-1:0{numlen}X} "
                f"[{','.join(f'0x{x:02X}' for x in self.mem[i : min(i + chunk_size, end_addr+1)])}]"
            )



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Unified Chip Design Platform - AMBA - AHB Multilayer.
"""

from logging import getLogger
from typing import ClassVar

import ucdp as u
from ucdp_addr import AddrMaster, AddrMatrix, AddrRef, AddrSlave

from . import types as t

LOGGER = getLogger(__name__)


class Master(AddrMaster):
    """
    Master.
    """

    proto: t.AmbaProto
    """Protocol Version."""


class Slave(AddrSlave):
    """
    Slave.
    """

    proto: t.AmbaProto
    """Protocol Version."""


class AhbFsmMlType(u.AEnumType):
    """
    FSM Type for AHB Multilayer.
    """

    keytype: u.UintType = u.UintType(3)
    title: str = "AHB ML FSM Type"
    comment: str = "AHB ML FSM Type"

    def _build(self):
        self._add(0, "idle", "No transfer")
        self._add(1, "transfer", "Transfer")
        self._add(2, "transfer_finish", "Transfer Finish (wait for HREADY)")
        self._add(3, "transfer_wait", "Transfer Wait for Grant")
        self._add(4, "error0", "Pre-Error (wait for HREADY)")
        self._add(5, "error1", "1st Error Cycle")
        self._add(6, "error2", "2ns Error Cycle")


class UcdpAhbMlMod(u.ATailoredMod, AddrMatrix):
    """
    AHB Multilayer.

    Multilayer.
    """

    filelists: ClassVar[u.ModFileLists] = (
        u.ModFileList(
            name="hdl",
            gen="full",
            filepaths=("$PRJROOT/{mod.topmodname}/{mod.modname}.sv"),
            template_filepaths=("ucdp_ahb_ml.sv.mako", "sv.mako"),
        ),
    )
    proto: t.AmbaProto = t.AmbaProto()
    """Default Protocol."""
    is_sub: bool = False
    """Full Address Decoding By Default."""
    addrwidth: int = 32
    datawidth: int = 32

    def _build(self):
        self.add_port(u.ClkRstAnType(), "main_i")

    def add_master(
        self,
        name: str,
        slavenames: u.Names | None = None,
        proto: t.AmbaProto | None = None,
        route: u.Routeable | None = None,
    ) -> Master:
        """
        Add master port named `name` connected to `route`.

        Args:
            name: Name or Pattern ('*' is supported)

        Keyword Args:
            slavenames: Names of slaves to be accessed by this master.
            proto: Protocol.
            route: port to connect this master to.
        """
        self.check_lock()
        proto = proto or self.proto
        master = Master(name=name, proto=proto)
        self._add_master(master, slavenames=slavenames)

        portname = f"ahb_mst_{name}_i"
        title = f"AHB Input {name!r}"
        self.add_port(t.AhbMstType(proto=proto), portname, title=title, comment=title)
        if route:
            self.con(portname, route)

        return master

    def add_slave(
        self,
        name: str,
        baseaddr=u.AUTO,
        size: u.Bytes | None = None,
        proto: t.AmbaProto | None = None,
        masternames: u.Names | None = None,
        route: u.Routeable | None = None,
        ref: u.BaseMod | str | None = None,
    ):
        """
        Add APB Slave.

        Args:
            name: Slave Name.

        Keyword Args:
            baseaddr: Base address, Next Free address by default. Do not add address space if `None`.
            size: Address Space.
            proto: AMBA Protocol Selection.
            masternames: Names of masters to be accessed by this slave.
            route: APB Slave Port to connect.
            ref: Logical Module connected.
        """
        self.check_lock()
        proto = proto or self.proto
        slave = Slave(name=name, addrdecoder=self, proto=proto, ref=ref)
        self._add_slave(slave, masternames=masternames, baseaddr=baseaddr, size=size)

        portname = f"ahb_slv_{name}_o"
        title = f"AHB Output {name!r}"
        self.add_port(t.AhbSlvType(proto=proto), portname, title=title, comment=title)
        if route:
            self.con(portname, route)

        return slave

    def _build_dep(self):
        self._check_masters_slaves()
        self.add_type_consts(t.AhbTransType())
        self.add_type_consts(t.AhbRespType())
        self.add_type_consts(t.AhbSizeType())
        self.add_type_consts(t.AhbBurstType())
        self.add_type_consts(t.AhbWriteType())
        self.add_type_consts(AhbFsmMlType(), name="fsm", item_suffix="st")
        for master in self.masters:
            master_slaves = self._master_slaves[master.name]
            self.add_signal(AhbFsmMlType(), f"fsm_{master.name}_r", comment=f"Master {master.name!r} FSM")
            self.add_signal(u.BitType(), f"mst_{master.name}_new_xfer_s")
            self.add_signal(u.BitType(), f"mst_{master.name}_cont_xfer_s")
            self.add_signal(u.BitType(), f"mst_{master.name}_hready_s")
            self.add_signal(u.BitType(), f"mst_{master.name}_rqstate_s")
            self.add_signal(u.BitType(), f"mst_{master.name}_addr_err_s")
            for slave in master_slaves:
                self.add_signal(u.BitType(), f"mst_{master.name}_{slave}_sel_s")
                self.add_signal(u.BitType(), f"mst_{master.name}_{slave}_req_r")
                self.add_signal(u.BitType(), f"mst_{master.name}_{slave}_gnt_r")
            self.add_signal(u.BitType(), f"mst_{master.name}_gnt_s")

            mstp_type = self.ports[f"ahb_mst_{master.name}_i"].type_
            for subt in mstp_type.values():
                if (subt.orientation == u.BWD) or (subt.name == "hwdata"):
                    continue
                self.add_signal(subt.type_, f"mst_{master.name}_{subt.name}_s")
                self.add_signal(subt.type_, f"mst_{master.name}_{subt.name}_r")
            self.add_signal(t.AhbWriteType(), f"mst_{master.name}_hwrite_dph_r", comment="data-phase write indicator")
        for slave in self.slaves:
            slave_masters = self._slave_masters[slave.name]
            num_mst = len(slave_masters)
            for master in slave_masters:
                self.add_signal(u.BitType(), f"mst_{master}_{slave.name}_req_s")
                if num_mst > 1:
                    self.add_signal(u.BitType(), f"mst_{master}_{slave.name}_keep_s")
                    self.add_signal(u.BitType(), f"slv_{slave.name}_{master}_gnt_r")
                    self.add_signal(u.BitType(), f"slv_{slave.name}_{master}_sel_s")
                self.add_signal(u.BitType(), f"slv_{slave.name}_{master}_gnt_s")

    @staticmethod
    def build_top(**kwargs):
        """Build example top module and return it."""
        return UcdpAhbMlExampleMod()

    def _resolve_ref(self, ref: AddrRef) -> AddrRef:
        return self.parent.parser(ref)

    def get_overview(self) -> str:
        """Matrix Overview."""
        return AddrMatrix.get_overview(self)


class UcdpAhbMlExampleMod(u.AMod):
    """
    Just an Example Multilayer.

    >>> print(UcdpAhbMlExampleMod().get_inst('u_ml').get_overview())
     Master > Slave    ram    periph    misc
    ----------------  -----  --------  ------
          ext           X                X
          dsp           X       X
    <BLANKLINE>
    <BLANKLINE>
    Size: 3932320 KB
    <BLANKLINE>
    | Addrspace | Type     | Base       | Size                      | Attributes |
    | --------- | ----     | ----       | ----                      | ---------- |
    | reserved0 | Reserved | 0x0        | 536870912x32 (2 GB)       |            |
    | misc      | Slave    | 0x80000000 | 5888x32 (23 KB)           |            |
    | reserved1 | Reserved | 0x80005C00 | 469756160x32 (1834985 KB) |            |
    | ram       | Slave    | 0xF0000000 | 16384x32 (64 KB)          |            |
    | periph    | Slave    | 0xF0010000 | 16384x32 (64 KB)          |            |
    | misc      | Slave    | 0xF0020000 | 8192x32 (32 KB)           |            |
    | reserved2 | Reserved | 0xF0028000 | 67067904x32 (261984 KB)   |            |
    <BLANKLINE>
    """

    def _build(self):
        ml = UcdpAhbMlMod(self, "u_ml")
        ml.add_master("ext")
        ml.add_master("dsp")

        slv = ml.add_slave("ram", masternames=["ext", "dsp"])
        slv.add_addrrange(0xF0000000, size=2**16)

        slv = ml.add_slave("periph")
        slv.add_addrrange(0xF0010000, size="64kb")

        slv = ml.add_slave("misc")
        slv.add_addrrange(size="32k")
        slv.add_addrrange(0x80000000, size="23k")

        # slv = ml.add_slave("ext", masternames=["ext", "dsp"])
        # slv.add_addrrange(0x0, size=2**32)
        # slv.add_exclude_addrrange(0xF0000000, size=2**18)

        ml.add_interconnects("dsp", "periph")
        ml.add_interconnects("ext", "misc")



#
# MIT License
#
# Copyright (c) 2024 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""
Unified Chip Design Platform - AMBA - AHB2APB.
"""

from logging import getLogger
from typing import ClassVar

import ucdp as u
from humannum import bytes_
from icdutil import num
from ucdp_addr import AddrDecoder, AddrRef, AddrSlave
from ucdp_glbl.irq import LevelIrqType

from . import types as t

LOGGER = getLogger(__name__)


class Slave(AddrSlave):
    """
    Slave.
    """

    proto: t.AmbaProto
    """Protocol Version."""


class Ahb2ApbFsmType(u.AEnumType):
    """
    FSM Type for AHB to APB Bridge.
    """

    keytype: u.UintType = u.UintType(3)
    title: str = "AHB to APB FSM Type"
    comment: str = "AHB to APB FSM Type"
    writeopt: bool = False

    def _build(self):
        self._add(0, "idle", "No transfer")
        self._add(1, "apb_ctrl", "Control Phase")
        if self.writeopt:
            self._add(2, "apb_data_wr", "Optimized Write")
        self._add(3, "apb_data", "Data Phase")
        self._add(4, "ahb_finish", "Finish Phase")
        self._add(5, "ahb_err", "Error Phase")
        self._add(6, "ahb_busy_finish", "Finish w/ Busy")
        self._add(7, "ahb_busy_err", "Error w/ Busy")


class UcdpAhb2apbMod(u.ATailoredMod, AddrDecoder):
    """
    AHB to APB Bridge.

    >>> class Mod(u.AMod):
    ...     def _build(self):
    ...         ahb2apb = UcdpAhb2apbMod(self, "u_ahb2apb")
    ...         ahb2apb.add_slave("uart")
    ...         ahb2apb.add_slave("spi")

    >>> ahb2apb = Mod().get_inst("u_ahb2apb")
    >>> print(ahb2apb.get_overview())
    Size: 8 KB
    <BLANKLINE>
    | Addrspace | Type  | Base    | Size           | Attributes |
    | --------- | ----  | ----    | ----           | ---------- |
    | uart      | Slave | +0x0    | 1024x32 (4 KB) | Sub        |
    | spi       | Slave | +0x1000 | 1024x32 (4 KB) | Sub        |
    <BLANKLINE>
    """

    filelists: ClassVar[u.ModFileLists] = (
        u.ModFileList(
            name="hdl",
            gen="full",
            filepaths=("$PRJROOT/{mod.topmodname}/{mod.modname}.sv"),
            template_filepaths=("ucdp_ahb2apb.sv.mako", "sv.mako"),
        ),
    )

    proto: t.AmbaProto = t.AmbaProto()
    errirq: bool = False
    writeopt: bool = False
    is_sub: bool = True
    default_size: u.Bytes | None = 4096
    ahb_addrwidth: int = 32
    datawidth: int = 32

    def _build(self):
        self.add_port(u.ClkRstAnType(), "main_i")
        if self.errirq:
            title = "APB Error Interrupt"
            self.add_port(LevelIrqType(), "irq_o", title=title, comment=title)
        self.add_port(
            t.AhbSlvType(proto=self.proto, addrwidth=self.ahb_addrwidth, datawidth=self.datawidth), "ahb_slv_i"
        )

    def add_slave(
        self,
        name: str,
        baseaddr=u.AUTO,
        size: u.Bytes | None = None,
        proto: t.AmbaProto | None = None,
        route: u.Routeable | None = None,
        ref: u.BaseMod | str | None = None,
    ):
        """
        Add APB Slave.

        Args:
            name: Slave Name.

        Keyword Args:
            baseaddr: Base address, Next Free address by default. Do not add address space if `None`.
            size: Address Space.
            proto: AMBA Protocol Selection.
            route: APB Slave Port to connect.
            ref: Logical Module connected.
            auser: User vector if there is not incoming `auser` signal.
        """
        proto = proto or self.proto
        slave = Slave(name=name, addrdecoder=self, proto=proto, ref=ref)
        self.slaves.add(slave)
        size = bytes_(size or self.default_size)
        if baseaddr is not None and (size is not None or self.default_size):
            slave.add_addrrange(baseaddr, size)

        portname = f"apb_slv_{name}_o"
        title = f"APB Slave {name!r}"
        self.add_port(
            t.ApbSlvType(proto=proto, addrwidth=num.calc_unsigned_width(size - 1), datawidth=self.datawidth),
            portname,
            title=title,
            comment=title,
        )
        if route:
            self.con(portname, route)

        return slave

    def _check_slaves(self):
        if not self.slaves:
            LOGGER.error("%r: has no APB slaves", self)
        slvchk = []
        for aspc in self.addrmap:
            if aspc.name in slvchk:
                raise ValueError(f"Slave {aspc.name!r} has non-contiguous address range.")
            slvchk.append(aspc.name)

    def _build_dep(self):
        self._check_slaves()
        self.add_type_consts(t.AhbTransType())
        self.add_type_consts(t.ApbReadyType())
        self.add_type_consts(t.ApbRespType())
        self.add_type_consts(Ahb2ApbFsmType(writeopt=self.writeopt), name="fsm", item_suffix="st")
        self.add_signal(u.BitType(), "ahb_slv_sel_s")
        self.add_signal(u.BitType(), "valid_addr_s")
        self.add_signal(Ahb2ApbFsmType(), "fsm_r")
        self.add_signal(t.AhbReadyType(), "hready_r")
        if not self.errirq:
            self.add_signal(t.ApbRespType(), "hresp_r")
        rng_bits = [num.calc_unsigned_width(aspc.size - 1) for aspc in self.addrmap]
        self.add_signal(t.ApbAddrType(max(rng_bits)), "paddr_r")
        self.add_signal(t.ApbWriteType(), "pwrite_r")
        self.add_signal(t.ApbDataType(self.datawidth), "pwdata_s")
        self.add_signal(t.ApbDataType(self.datawidth), "pwdata_r")
        self.add_signal(t.ApbDataType(self.datawidth), "prdata_s")
        self.add_signal(t.ApbDataType(self.datawidth), "prdata_r")
        self.add_signal(t.ApbEnaType(), "penable_r")
        self.add_signal(t.ApbReadyType(), "pready_s")
        self.add_signal(t.ApbRespType(), "pslverr_s")
        for aspc in self.addrmap:
            self.add_signal(t.ApbSelType(), f"apb_{aspc.name}_sel_s")
            self.add_signal(t.ApbSelType(), f"apb_{aspc.name}_sel_r")
        if self.errirq:
            self.add_signal(LevelIrqType(), "irq_r")

    def get_overview(self):
        """Overview."""
        return self.addrmap.get_overview(minimal=True)

    @staticmethod
    def build_top(**kwargs):
        """Build example top module and return it."""
        return UcdpAhb2apbExampleMod()

    def _resolve_ref(self, ref: AddrRef) -> AddrRef:
        return self.parent.parser(ref)


class UcdpAhb2apbExampleMod(u.AMod):
    """
    Just an Example.
    """

    def _build(self):
        class SecIdType(t.ASecIdType):
            def _build(self):
                self._add(0, "apps")
                self._add(2, "comm")
                self._add(5, "audio")

        amba5 = t.AmbaProto(name="amba5", secidtype=SecIdType(default=2))

        for errirq in (False, True):
            for proto in (t.AMBA3, amba5):
                name = f"u_ahb2apb_{proto.name}_errirq{errirq}".lower()
                ahb2apb = UcdpAhb2apbMod(self, name, proto=proto, errirq=errirq)
                ahb2apb.add_slave("default")
                ahb2apb.add_slave("slv3", proto=t.AMBA3)
                ahb2apb.add_slave("slv5", proto=amba5)

        ahb2apb = UcdpAhb2apbMod(self, "u_odd", ahb_addrwidth=27, errirq=False)
        ahb2apb.add_slave("foo")
        ahb2apb.add_slave("bar", size="1KB")
        ahb2apb.add_slave("baz", size="13kB")
        # slv.add_addrrange(size="3kB")
